/*
 * IDLflex -- A flexible, generic code generator for CORBA IDL
 *
 * Copyright (C) 2000  Hans Reiser, AspectIX Research Team
 *                     University of Erlangen-Nuernberg, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You find a copy of the license in the LICENSE.txt file in the
 *  root of the software distribution; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Please read the README.txt file first.
 *
 *  $Id: IDLflexMain.java,v 1.5 2001/09/07 09:20:42 reiser Exp $
 */

package org.aspectix.IDLflex;

import org.omg.CORBA.*;
import org.aspectix.IDLflex.*;
import org.aspectix.IDLflex.IRObj.*;
import java.util.*;
import java.io.*;


/** IDLflexMain implements the main method, which is responsible for the
 *  initialization of the CORBA environment and the Interface Repository.
 *  External modules are responsible for vendor-specific stuff 
 *  (IR_JavaORB, IR_ORBacus).
 *
 *  Most of the code is used for parsing command line arguments and
 *  setting up the IR connection.
 */
final public class IDLflexMain {
    public static int verbose=0;
    static org.omg.CORBA.Object obj = null;
    public static org.omg.CORBA.Repository ir = null;
    public static IDLDefCompiler c;
    private static IRBase irbase=null;
    static Vector ORBArguments;
    static {
	ORBArguments = new Vector();
    }

    private static void printVersionAndExit()
    {
	System.out.println("IDLflex version "+Version.VERSIONSTRING);
	System.exit(0);
    }

    /** returns a reference to an external interface repository.
     *
     *  if ior_arg is not null, an ORB paramter 
     *  "-ORBservice InterfaceRepository <ior_arg>"
     *  will be passed to the ORB initialization.
     *
     *  Calls orb.resolve_initial_references("InterfaceRepository") to get
     *  the apropriate reference.
     */
    public static Repository resolve_ir(String ior_arg) {
	String [] args;
	Repository ir;

	// 1. ORB-INIT
	if(verbose>0)
	    System.out.println(">> IDLflexMain: Initializing CORBA ORB...");
	if(ior_arg!=null && !ior_arg.equals("")) {
	    ORBArguments.addElement("-ORBservice");
	    ORBArguments.addElement("InterfaceRepository");
	    ORBArguments.addElement(ior_arg);
	}
	args = new String[ORBArguments.size()];
	for(int i=0; i<args.length; i++) {
	    args[i] = (String)ORBArguments.elementAt(i);
	}
	org.omg.CORBA.ORB orb=org.omg.CORBA.ORB.init(args,null);
	
	// 2. Resolve the InterfaceRepository
	if(verbose>0)
	    System.out.println(">> IDLflexMain: Resolving the Interface Repository");
	try {
	    obj = orb.resolve_initial_references("InterfaceRepository");
	    ir = org.omg.CORBA.RepositoryHelper.narrow(obj);
	    if(ir==null) throw new org.omg.CORBA.ORBPackage.InvalidName();
	    return ir;
	}
	catch ( org.omg.CORBA.ORBPackage.InvalidName name ) {
	    System.err.println("Unable to resolve the Interface Repository");
	    System.exit(1);
	}
	return null;
    }

    /** Uses IR_JavaORB or IR_ORBacus to automatically create and fill
     *  an interface repository.
     *
     * @exception IDLflexException  if the external interface could not
     *     be started or filled with the IDL data.
     */
    public static Repository getExternalIR( String argsFileName ) 
    throws IDLflexException {
	if(verbose>0)
	    System.out.println(">> IDLflexMain: Creating interface repository"
			       +" from file "+argsFileName);
	try {
	    Class c = Class.forName("org.aspectix.IDLflex.IR_JavaORB");
	    irbase = (IRBase)(c.newInstance());
	    ir = irbase.create_repository( argsFileName );
	    if(verbose>0)
		System.out.println(">> IDLflexMain: Using JavaORB Interface Repository");
	    return ir;
	} 
	catch(Throwable e) {
	    e.printStackTrace();
	}
	try {
	    Class c = Class.forName("org.aspectix.IDLflex.IR_ORBacus");
	    irbase = (IRBase)(c.newInstance());
	    ir = irbase.create_repository( argsFileName );
	    if(verbose>0) 
		System.out.println(">> IDLflexMain: Using ORBacus Interface Repository");
	    return ir;
	}
	catch(Throwable e) {
	}
	throw new IDLflexException("Cannot feed "+argsFileName+" to internal repository:\nUnable to use either JavaORB or ORBacus IR tool");
    }



    public static String argsFileName = null;
    public static String argsMapping = null;
    public static class _Attribute {
	String[] names;
	String[] values;
	_Attribute(Vector attr) {
	    names = new String[attr.size()];
	    values = new String[attr.size()];
	    for(int i=0; i<attr.size(); i++) {
		String data = ((String)attr.elementAt(i));
		int pos = data.indexOf("=");
		if(pos>0) {
		    names[i] = data.substring(0,pos);
		    values[i] = data.substring(pos+1);
		} else {
		    names[i] = data;
		    values[i] = "";
		}
		if(verbose>1)
		    System.err.println("defined: "+names[i]+" = "+values[i]);
	    }
	}
	boolean isDefined(String s) {
	    int len=names.length;
	    for(int i=0; i<len; i++) {
		if(names[i].equals(s)) return true;
	    }
	    return false;
	}
	String getDefined(String s) {
	    int len=names.length;
	    for(int i=0; i<len; i++) {
		if(names[i].equals(s)) return values[i];
	    }
	    return null;
	}
    }
    public static _Attribute argsAttributes;
    public static String[] argsInclude;
    public static String[] argsExclude;

    /**	Command line parsing.
     */
    public static int process_args( String args[] ) {
	int state=0;
	Vector Attributes = new Vector();
	Vector Include = new Vector();
	Vector Exclude = new Vector();

	for(int i=0; i<args.length; i++) {
	    String data=args[i];
	    if(state==0) {
		if(args[i].length()>2) data=args[i].substring(2);
		else data="";
		if(args[i].startsWith("-D")) state=1;
		else if(args[i].startsWith("-I")) state=2;
		else if(args[i].startsWith("-X")) state=3;
		else if(args[i].startsWith("-m")) state=4;
		else if(args[i].startsWith("-V")) {
		    printVersionAndExit();
		}
		else if(args[i].startsWith("-v")) verbose++; 
		else if(args[i].equals("--")) { state=5; }
		else if(args[i].startsWith("-")) {
		    System.err.println("Illegal argument: "+args[i]);
		    break; //return -1;
		} else {
		    if(argsFileName==null) argsFileName=args[i];
		    else {
			System.err.println("Invalid argument");
		    }
		}
		if(data.equals("")) continue;
	    }
	    if(state==1) {
		Attributes.addElement(data); state=0;
	    } else if(state==2) { 
		Include.addElement(data); state=0;
	    } else if(state==3) { 
		Exclude.addElement(data); state=0;
	    } else if(state==4) {
		if(argsMapping==null) argsMapping=data;
		else {
		    System.err.println("Duplicate Option '-m'");
		    break; //return -1;
		}
		state=0;
	    } else if(state==5) {
		ORBArguments.addElement(data);
	    }
	}
	if(state!=0 && state!=5) {
	    System.err.println("Option missing argument:"+args[args.length-1]);
	    //return -1;
	}
	argsAttributes = new _Attribute(Attributes);

	argsInclude = new String[Include.size()];
	for(int i=0; i<Include.size(); i++)
	    argsInclude[i] = (String)Include.elementAt(i);
	argsExclude = new String[Exclude.size()];
	for(int i=0; i<Exclude.size(); i++)
	    argsExclude[i] = (String)Exclude.elementAt(i);
	if(argsMapping==null)
	    argsMapping="JavaMapping.xml";
	return 0;
    }

    /** main method: 
     *<ul>
     *  <li>parse command line arguments
     *  <li>connect to interface repository
     *  <li>locate XML mapping document
     *  <li>initialize the compiler 
     *    (-> IDL object representation, XML parser tree)
     *  <li>run the compiler
     *</ul>
     */
    public static void main( String args[] )
    {
        // IR_ORBacus has to 'kill' external IR on exit...
	//	java.lang.Runtime.runFinalizersOnExit(true);

	// 1. Parse command line

	String IDLflexHome = java.lang.System.getProperty("IDLflexHome",".");

	int res = process_args( args );
	if(res<0) {
	    System.err.println("Usage: ");
	    System.err.println("java IDLflex.IDLflexMain [-D<defines>] [-I<includes>] [-X<excludes>] [-m<Mapping.xml>] [<filename>]");
	    System.exit(1);
	}

	if(argsFileName!=null) {
	    try {
		ir = getExternalIR( argsFileName );
	    }
	    catch(IDLflexException e) {
		e.printStackTrace();
		System.exit(1);
	    }
	} else {
	    ir = resolve_ir( argsAttributes.getDefined("ORBacusIRRef") );
	}

	// 3. Read XML configuration file and create compiler object

	if(verbose>0)
	    System.out.println(">> IDLflexMain: Parsing the Java Mapping specification...");
	c=null;
	String mapName; File mapFile;
	mapName = argsMapping; mapFile=new File(mapName);
	if(!mapFile.exists()) {
	    mapName = IDLflexHome+"/"+argsMapping; mapFile = new File(mapName);
	    if(!mapFile.exists()) {
		mapName = IDLflexHome+"/mapping/"+argsMapping;
		mapFile = new File(mapName);
		if(!mapFile.exists()) {
		    System.err.println("Error: Mapping-File not found: "+
				       argsMapping);
		    System.exit(1);
		}
	    }
	}
	
	// 3b) Create the compiler
	if(verbose>0)
	    System.out.println(">> IDLflexMain: Initializing the compiler object...");
	try {
	    c = new IDLDefCompiler(mapName);
	} catch(Exception e) {
	    e.printStackTrace();
	    System.exit(1);
	}
	
	// 4. parse the IR tree
	if(verbose>0)
	    System.out.println(">> IDLflexMain: Parsing the IR Object tree...");
	IDLObject iobj;
	try {
	    iobj = ObjectCreator.get(ir);
	} catch(Exception e) {
	    e.printStackTrace();
	    System.exit(1);
	}

	// 4. and run the compiler...
	if(verbose>0)
	    System.out.println(">> IDLflexMain: Running the compiler...");
	try {
	    String root=c.getRootCompiler();
	    c.generateCode(root, ObjectCreator.get(ir));
	} catch( Exception e ) {
	    System.err.println("\nRepositoryDefCompiler failed:");
	    e.printStackTrace();
	    System.exit(1);
	}

	if(irbase!=null) { irbase.cleanup(); irbase=null; }
	if(verbose>0)
	    System.out.println("done.");
	System.exit(1);
    }

    protected void finalize() throws Throwable
    {
	if(irbase!=null) irbase.cleanup();
    }
}
