/*
 * IDLflex -- A flexible, generic code generator for CORBA IDL
 *
 * Copyright (C) 2000  Hans Reiser, AspectIX Research Team
 *                     University of Erlangen-Nuernberg, Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License.
 * The full text of the license can be found in the LICENSE.txt file.
 *
 * $Id: ObjectCreator.java,v 1.4 2001/06/15 09:08:31 fzhauck Exp $
 */

package org.aspectix.IDLflex.IRObj;

import org.omg.CORBA.*;
import java.util.Hashtable;
import org.aspectix.IDLflex.IDLflexException;


/** This class is used to create IDL objects from Interface Repository
 *  objects. Depending on the type of the IR definition, the
 *  corresponding IDL object is created.  
 */
public class ObjectCreator {

    /** Hashtable for created IDLObjects.
     */
    static Hashtable ht;
    static {
	ht = new Hashtable();
    }

    /** Translate IDL object ID into a java.lang.Object used in lookup
     *  hashtable.
     */
    private static java.lang.Object getID(IRObject iro) {
	try {
	    Contained c=ContainedHelper.narrow(iro);
	    return c.id();
	} catch(Exception e) {};
	try {
	    IDLType idl = IDLTypeHelper.narrow(iro);
	    return idl.type();
	} catch(Exception e) {};
	return iro;
    }

    /** Method get tries to return a reference to an already existing object.
     *  if no such object is found, the behavior depends on the
     *  paramter create: if it is set, a new object will be created,
     *  otherwise null is returned.
     */
    public static IDLObject get(IRObject iro, boolean create) 
	throws IDLflexException 
    {
	IDLObject obj = (IDLObject)ht.get( getID(iro) );
	if(obj!=null) {
	    return obj;
	} else {
	    if(create) {
		//System.err.println("Creating: "+getID(iro) );
		IDLObject o = create_obj(iro);
		if(o!=null) {
		    // First we have to register the object. Some loop
		    // in the definitions might have to lookup the 
		    // current object in the hash table in its
		    // init_members function.
		    ht.put( getID(iro), o);
		    o.init_members();
		} else {
		    System.err.println("ObjectCreator: create failed on "+iro);
		}
		return o;
	    } else {
		return null;
	    }
	}
    }

    /**	Just an alias to
     *	{@link #get(org.omg.CORBA.IRObject,boolean) get}(iro, true); 
     */
    public static IDLObject get(IRObject iro) 
	throws IDLflexException
    {
	if(iro==null) 
	    throw new IDLflexException("ObjectCreator: reference object is null");
	return get(iro, true);
    }

    /** Creates an IDLObject according to the type of the IRObject.
     */
    private static IDLObject create_obj(IRObject iro) {
	switch(iro.def_kind().value()) {
	case DefinitionKind._dk_Attribute:
	    return new AttributeObj(iro);
	case DefinitionKind._dk_Constant:
	    return new ConstantObj(iro);
	case DefinitionKind._dk_Exception:
	    return new ExceptionObj(iro);
	case DefinitionKind._dk_Interface:
	    return new InterfaceObj(iro);
	case DefinitionKind._dk_Module:
	    return new ModuleObj(iro);
	case DefinitionKind._dk_Operation:
	    return new OperationObj(iro);
	case DefinitionKind._dk_Typedef:
	    return new TypedefObj(iro);
	case DefinitionKind._dk_Alias:
	    return new AliasObj(iro);
	case DefinitionKind._dk_Struct:
	    return new StructObj(iro);
	case DefinitionKind._dk_Union:
	    return new UnionObj(iro);
	case DefinitionKind._dk_Enum:
	    return new EnumObj(iro);
	case DefinitionKind._dk_Primitive:
	    {
		PrimitiveDef pd = PrimitiveDefHelper.narrow(iro);
		// String Obj: refObj, IsBound, IsWide
		if( pd.kind()==PrimitiveKind.pk_string)
		    return new StringObj(iro, false, false);
		if( pd.kind()==PrimitiveKind.pk_wstring)
		    return new StringObj(iro, false, true);
		return new PrimitiveObj(iro);
	    }
	case DefinitionKind._dk_String:
	    return new StringObj(iro, true, false);
	case DefinitionKind._dk_Wstring:
	    return new StringObj(iro, true, true);
	case DefinitionKind._dk_Sequence:
	    return new SequenceObj(iro);
	case DefinitionKind._dk_Array:
	    return new ArrayObj(iro);
	case DefinitionKind._dk_Repository:
	    return new ContainerObj(iro);
	case DefinitionKind._dk_Fixed:
	    return new FixedObj(iro);
	default:
	    return null;
	}
    }
}
