/*
 * IDLflex -- A flexible, generic code generator for CORBA IDL
 *
 * Copyright (C) 2000  Hans Reiser, AspectIX Research Team
 *                     University of Erlangen-Nuernberg, Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License.
 * The full text of the license can be found in the LICENSE.txt file.
 *
 * $Id: UnionObj.java,v 1.4 2001/09/07 09:20:45 reiser Exp $
 */

package org.aspectix.IDLflex.IRObj;

import java.lang.Object;
import org.omg.CORBA.*;
import org.aspectix.IDLflex.IDLflexException;
import java.util.Enumeration;
import java.util.Vector;

/** implementation object of an IDL union.
 *
 * the list of contained definition elements can be accessed in two ways:
 * either a flat list (with some multiple entries which just differ by 
 * their labels) can be obtained, or a list with unique definition
 * entries, each of them having a sublist with their labels (if they
 * have multiple labels)
 *
 *<ul>
 * <li><em>getAttribute("needsdef")</em>
 *     true, if the union requires an implicit default declaration, i.e.
 *     if the existing labels do not cover completely the domain of the
 *     discriminator data type.
 * <li><em>getName/getValue("defdiscr")</em>
 *     returns the smalles possible value (index number) of the default
 *     label, either as a String or as an Long object.
 * <li><em>getValue("discrtype")</em>
 *     returns the TypeCode of the discriminator
 * <li><em>getContent("DISCR")</em>
 *     returns the IDL object defining the discriminator data type
 * <li><em>getContent("RDISCR")</em>
 *     returns the IDL object defining the discriminator data type.
 *     If the discriminator type is an alias, the type referenced by
 *     the alias is returned, i.e. the alias is resolved.
 * <li><em>getContentList("ELEMENTS")</em>
 *     returns the flat list of all contained UnionMemberObj objects
 * <li><em>getContentList("UNIQUE")</em>
 *     returns a list of all unique UnionMemberObj defining union
 *     entries.
 *</ul>
 */
public class UnionObj extends ContainerObj {
    Vector union_flat;
    Vector union_unique;
    IDLObject discriminator;
    IDLObject r_discriminator;  // alias-resolved discriminator 
    boolean implicit_default;
    long def_discr_value;
    TypeCode type;


    public UnionObj(org.omg.CORBA.IRObject iro) {
	super( iro ); 
    }

    /** extract the value of union label from an Any */
    static long getLabelNumber(Any a) throws BAD_OPERATION 
    { 
	try { return (long)a.extract_longlong(); } catch(BAD_OPERATION e) {};
	try { return (long)a.extract_ulonglong(); } catch(BAD_OPERATION e) {};
	try { return (long)a.extract_long(); } catch(BAD_OPERATION e) {};
	try { return (long)a.extract_ulong(); } catch(BAD_OPERATION e) {};
	try { return (long)a.extract_short(); } catch(BAD_OPERATION e) {};
	try { return (long)a.extract_ushort(); } catch(BAD_OPERATION e) {};
	try { return (long)a.extract_char(); } catch(BAD_OPERATION e) {};
	try { return (long)a.extract_wchar(); } catch(BAD_OPERATION e) {};
	try { return (long)a.extract_octet(); } catch(BAD_OPERATION e) {};
	try { return a.extract_boolean()?1:0; } catch(BAD_OPERATION e) {};
	try { return (long)EnumObj.valueFromAny(a); } catch(Exception e) {
	    e.printStackTrace();
	    //System.err.println(e);
	}
	System.err.println("getLabelNumber: kind="+a.type().kind().value());
	throw new BAD_OPERATION();
    }

    /** compute the maximal value of the discriminator data type */
    private int value_maxval(IDLType t) {
	TypeCode tc = t.type();
	while(tc.kind() == TCKind.tk_alias) {  // remove aliases
	    t = (AliasDefHelper.narrow(t)).original_type_def();
	    tc = t.type();
	}
	if(tc.kind()==TCKind.tk_short||tc.kind()==TCKind.tk_ushort)
	    return 65535;
	if( tc.kind()==TCKind.tk_long||tc.kind()==TCKind.tk_ulong||
	    tc.kind()==TCKind.tk_longlong||tc.kind()==TCKind.tk_ulonglong )
	    return 2000000000; // Not the correct value, but this doesn't
	                       // matter.
	if(tc.kind()==TCKind.tk_boolean) return 1;
	if(tc.kind()==TCKind.tk_char||tc.kind()==TCKind.tk_octet)
	    return 255;
	if(tc.kind()==TCKind.tk_wchar) return 65535;
	if(tc.kind()==TCKind.tk_enum) {
	    try {
		return tc.member_count()-1;
	    } catch(Exception e) {};
	}
	System.err.println("UnionObj: invalid typecode for discriminator");
	System.err.println("TCKind: "+tc.kind()+" -- "+tc.kind().value());
	return 999999999;
    }
    
    public void init_members() throws IDLflexException {
	super.init_members();
	if( referenceObject==null ) return;

	UnionDef ud = UnionDefHelper.narrow(referenceObject);
	UnionMember[] umlist = ud.members();

	discriminator = ObjectCreator.get(ud.discriminator_type_def());
	type = ud.discriminator_type();
	r_discriminator = discriminator;
	while(r_discriminator.is_a("AliasObj")) {
	    r_discriminator = r_discriminator.getContent("RESOLVE");
	}
	
	// members in a flat structure and structured by names...
	union_flat = new Vector();
	union_unique = new Vector();

	int unique_cnt=0, same_cnt=0;
	for(int i=0; i<umlist.length; i++) {
	    UnionMemberObj uobj = new UnionMemberObj( umlist, i, 0 );
	    uobj.dad = this;
	    uobj.init_members();
	    uobj.defined_in = this.defined_in;
	    union_flat.addElement(uobj);

	    same_cnt++;
	    if(i==umlist.length-1 ||
	       !umlist[i].name.equals(umlist[i+1].name)) {
		// last entry of a name!
		UnionMemberObj uunique = new 
		    UnionMemberObj(umlist, i-same_cnt+1, same_cnt);
		uunique.dad = this;
		uunique.init_members();
		uunique.defined_in = this.defined_in;
		union_unique.addElement(uunique);
		unique_cnt++;
		same_cnt=0;
	    }
	}

	{
	// if there is an explicit default, we dont need an implicit one.
	    boolean explicit=false;
	    Enumeration en = union_unique.elements();
	    while(en.hasMoreElements()) {
		UnionMemberObj umem = (UnionMemberObj)en.nextElement();
		try {
		    if(umem.getAttribute("isdefault")) {
			explicit=true; break;
		    }
		} catch(Exception e) {
		    System.err.println(e);
		}
	    }
	    if(explicit)
		implicit_default=false;
	    else {
		// check if we need an implicit default modifier branch
		// i.e. umlist.length < value_range(discriminator)
		if( umlist.length <= value_maxval( ud.discriminator_type_def() ) )
		    implicit_default=true; 
		else
		    implicit_default=false;
	    }
	}

	// set def_discr_value to first unused entry! 
	// probably this could be done more efficient!
	int max=value_maxval( ud.discriminator_type_def() );
    outer:
	for(def_discr_value=0; def_discr_value<max; def_discr_value++) {
	    Enumeration en = union_flat.elements();
	    while(en.hasMoreElements()) {
		UnionMemberObj umem =  (UnionMemberObj)en.nextElement();
		try {
		    if( umem.getAttribute("isdefault") ) 
			continue;
		    if( getLabelNumber( umem.label ) == def_discr_value )
			continue outer;
		} catch(IDLflexException e) {
		    System.err.println("in UnionObj-init: "+e);
		    break outer;
		}
	    }
	    break;
	}
    }

    public boolean getAttribute(String spec) throws IDLflexException {
	if(spec.equalsIgnoreCase("needsdef")) 
	    return implicit_default;
	return super.getAttribute(spec);
    }

    // return default discriminator value numerically
    public String getName(String spec) throws IDLflexException 
    {
	if(spec.equalsIgnoreCase("defdiscr")) 
	    return ""+def_discr_value;
	return super.getName(spec);
    }

    public Object getValue(String spec) throws IDLflexException 
    {
	if(spec.equalsIgnoreCase("defdiscr")) 
	    return new Long(def_discr_value);
	else if(spec.equalsIgnoreCase("discrtype")) 
	    return type;
	return super.getValue(spec);
    }

    public IDLObject getContent(String spec) throws IDLflexException {
	if(spec.equalsIgnoreCase("DISCR")) 
	    return discriminator;
	else if(spec.equalsIgnoreCase("RDISCR")) 
	    return r_discriminator;
	else return super.getContent(spec);
    }

    public IDLObject[] getContentList(String spec) throws IDLflexException {
	if(spec.equalsIgnoreCase("ELEMENTS")) 
	    return VectorToIDLObjects(union_flat);
	else if(spec.equalsIgnoreCase("UNIQUE")) 
	    return VectorToIDLObjects(union_unique);
	else return super.getContentList(spec);
    }

    public boolean is_a(String spec, boolean inherit) {
	if(spec.equalsIgnoreCase("UnionObj")) return true;
	if(inherit) return super.is_a(spec, inherit);
	else return false;
    }
}
